import httpx

from typing import Dict

import config
import mods
import log

logger = log.setup_logger(__name__)

MODEL_TO_BACKEND: Dict[str, str] = {}

async def list_downstream():
  logger.debug("Listing downstream models")

  all_models = []

  for url, key in zip(
    config.BOOST_APIS,
    config.BOOST_KEYS,
  ):
    try:
      endpoint = f"{url}/models"
      headers = {
        "Authorization": f"Bearer {key}",
        "Content-Type": "application/json",
      }

      logger.debug(f"Fetching models from '{endpoint}'")

      async with httpx.AsyncClient() as client:
        response = await client.get(endpoint, headers=headers)
        response.raise_for_status()
        json = response.json()
        models = json.get("data", [])

        logger.debug(f"Found {len(models)} models at '{endpoint}'")
        all_models.extend(models)

        for model in models:
          MODEL_TO_BACKEND[model["id"]] = url

    except Exception as e:
      logger.error(f"Failed to fetch models from {endpoint}: {e}")

  return all_models


def get_proxy_model(module, model: dict) -> Dict:
  return {
    **model,
    "id": f"{module.ID_PREFIX}-{model['id']}",
    "name": f"{module.ID_PREFIX} {model['id']}",
  }


def resolve_proxy_model(model_id: str) -> Dict:
  parts = model_id.split("-")
  if parts[0] in mods.registry:
    return "-".join(parts[1:])
  return model_id


def resolve_proxy_module(model_id: str) -> Dict:
  parts = model_id.split("-")
  if parts[0] in mods.registry:
    return parts[0]
  return None


def resolve_request_config(body: Dict) -> Dict:
  model = body.get("model")
  messages = body.get("messages")
  params = {k: v for k, v in body.items() if k not in ["model", "messages"]}

  if not model:
    raise ValueError("Unable to proxy request without a model specifier")

  proxy_model = resolve_proxy_model(model)
  proxy_module = resolve_proxy_module(model)
  proxy_backend = MODEL_TO_BACKEND.get(proxy_model)

  logger.debug(
    f"Resolved proxy model: {proxy_model}, proxy module: {proxy_module}, proxy backend: {proxy_backend}"
  )

  proxy_key = config.BOOST_KEYS[
    config.BOOST_APIS.index(proxy_backend)]

  return {
    "url": proxy_backend,
    "headers":
      {
        "Authorization": f"Bearer {proxy_key}",
        "Content-Type": "application/json",
      },
    "model": proxy_model,
    "params": params,
    "messages": messages,
    "module": proxy_module,
  }

def is_title_generation_task(llm: 'LLM'):
  # TODO: Better way to identify?
  return llm.chat.has_substring("3-5 word title")



from typing import Optional, AsyncGenerator
import traceback

import json
import asyncio
import time
import httpx

from config import INTERMEDIATE_OUTPUT
import chat as ch
import log
import format
import mods

logger = log.setup_logger(__name__)

BOOST_PARAM_PREFIX="@boost_"


class LLM:
  url: str
  headers: dict

  model: str
  params: dict
  boost_params: dict
  module: str

  queue: asyncio.Queue
  is_streaming: bool
  is_final_stream: bool

  cpl_id: int

  def __init__(self, **kwargs):
    self.url = kwargs.get('url')
    self.headers = kwargs.get('headers', {})

    self.model = kwargs.get('model')
    self.split_params(kwargs.get('params', {}))

    self.chat = self.resolve_chat(**kwargs)
    self.messages = self.chat.history()

    self.module = kwargs.get('module')

    self.queue = asyncio.Queue()
    self.is_streaming = False
    self.is_final_stream = False

    self.cpl_id = 0

  @property
  def chat_completion_endpoint(self):
    return f"{self.url}/chat/completions"

  def split_params(self, params: dict):
    self.params = {
      k: v for k, v in params.items() if not k.startswith(BOOST_PARAM_PREFIX)
    }
    self.boost_params = {
      k[len(BOOST_PARAM_PREFIX):]: v for k, v in params.items() if k.startswith(BOOST_PARAM_PREFIX)
    }

  def generate_system_fingerprint(self):
    return "fp_boost"

  def generate_chunk_id(self):
    return f"chatcmpl-{++self.cpl_id}"

  def get_response_content(self, response):
    return response['choices'][0]['message']['content']

  def get_chunk_content(self, chunk):
    return chunk["choices"][0]["delta"]["content"]

  def parse_chunk(self, chunk):
    if isinstance(chunk, dict):
      return chunk

    if isinstance(chunk, bytes):
      chunk = chunk.decode('utf-8')

    chunk_str = chunk.split("\n")[0]
    if chunk_str.startswith("data: "):
      chunk_str = chunk_str[6:]

    try:
      return json.loads(chunk_str)
    except json.JSONDecodeError:
      logger.error(f"Failed to parse chunk: {chunk_str}")
      return {}

  def output_from_chunk(self, chunk):
    return {
      "id": chunk["id"],
      "object": "chat.completion",
      "created": chunk["created"],
      "model": self.model,
      "system_fingerprint": self.generate_system_fingerprint(),
      "choices":
        [
          {
            "index": choice["index"],
            "message":
              {
                "role": choice["delta"].get("role", "assistant"),
                "content": choice["delta"].get("content", "")
              },
            "finish_reason": None
          } for choice in chunk["choices"]
        ],
      "usage": {
        "prompt_tokens": 0,
        "completion_tokens": 0,
        "total_tokens": 0
      }
    }

  def chunk_from_message(self, message: str):
    now = int(time.time())

    return {
      "id":
        self.generate_chunk_id(),
      "object":
        "chat.completion.chunk",
      "created":
        now,
      "model":
        self.model,
      "system_fingerprint":
        self.generate_system_fingerprint(),
      "choices":
        [
          {
            "index": 0,
            "delta": {
              "role": "assistant",
              "content": message
            },
            "finish_reason": None
          }
        ]
    }

  def chunk_to_string(self, chunk):
    if isinstance(chunk, dict):
      chunk = f"data: {json.dumps(chunk)}\n\n"

    return chunk

  async def serve(self):
    logger.debug('Serving boosted LLM...')

    if self.module is None:
      logger.debug("No module specified")
      return self.stream_chat_completion()

    mod = mods.registry.get(self.module)

    if mod is None:
      logger.error(f"Module '{self.module}' not found.")
      return

    async def apply_mod():
      logger.debug(f"Applying '{self.module}' to '{self.model}'")
      try:
        self.chat.llm = self
        await mod.apply(chat=self.chat, llm=self)
      except Exception as e:
        logger.error(f"Failed to apply module '{self.module}': {e}")
        for line in traceback.format_tb(e.__traceback__):
          logger.error(line)

      logger.debug(f"'{self.module}' application complete for '{self.model}'")
      await self.emit_done()

    asyncio.create_task(apply_mod())
    return self.response_stream()

  async def generator(self):
    self.is_streaming = True

    while self.is_streaming:
      chunk = await self.queue.get()

      if chunk is None:
        break
      yield chunk

  async def response_stream(self):
    async for chunk in self.generator():
      # Final stream is always passed back as
      # that's the useful payload of a given iteration
      if INTERMEDIATE_OUTPUT.value or self.is_final_stream:
        yield chunk

  async def emit_status(self, status):
    await self.emit_message(format.format_status(status))

  async def emit_message(self, message):
    await self.emit_chunk(self.chunk_from_message(message))

  async def emit_chunk(self, chunk):
    await self.queue.put(self.chunk_to_string(chunk))

  async def emit_done(self):
    await self.queue.put(None)
    self.is_streaming = False

  async def stream_final_completion(self, **kwargs):
    self.is_final_stream = True
    return await self.stream_chat_completion(**kwargs)

  async def stream_chat_completion(self, **kwargs):
    chat = self.resolve_chat(**kwargs)

    logger.debug(
      f"Streaming Chat Completion for '{self.chat_completion_endpoint}"
    )

    if chat is None:
      chat = self.chat

    result = ""

    async with httpx.AsyncClient(timeout=None) as client:
      async with client.stream(
        "POST",
        self.chat_completion_endpoint,
        headers=self.headers,
        json={
          "model": self.model,
          "messages": chat.history(),
          **self.params,
          "stream": True,
        }
      ) as response:
        async for chunk in response.aiter_bytes():
          parsed = self.parse_chunk(chunk)
          content = self.get_chunk_content(parsed)
          result += content

          # We emit done after the module
          # application has completed
          if not '[DONE]' in f"{chunk}":
            await self.emit_chunk(chunk)

    return result

  async def chat_completion(self, **kwargs):
    chat = self.resolve_chat(**kwargs)
    should_resolve = kwargs.get("resolve", False)

    logger.debug(f"Chat Completion for '{self.chat_completion_endpoint}'")
    logger.debug(f"Chat: {chat}")
    if chat is None:
      chat = self.chat

    async with httpx.AsyncClient(timeout=None) as client:
      body = {
        "model": self.model,
        "messages": chat.history(),
        **self.params, "stream": False
      }
      response = await client.post(
        self.chat_completion_endpoint, headers=self.headers, json=body
      )
      result = response.json()
      if should_resolve:
        return self.get_response_content(result)
      return result

  async def consume_stream(self, stream: AsyncGenerator[bytes, None]):
    output_obj = None
    content = ""

    async for chunk_bytes in stream:
      chunk = self.parse_chunk(chunk_bytes)
      if output_obj is None:
        output_obj = self.output_from_chunk(chunk)
      chunk_content = self.get_chunk_content(chunk)
      content += chunk_content

    if output_obj:
      output_obj["choices"][0]["message"]["content"] = content

    return output_obj

  def resolve_chat(
    self,
    messages: Optional[list] = None,
    chat: Optional['ch.Chat'] = None,
    prompt: Optional[str] = None,
    **prompt_kwargs
  ):
    if chat is not None:
      return chat

    if messages is not None:
      return ch.Chat.from_conversation(messages)

    if prompt is not None:
      message = prompt.format(**prompt_kwargs)
      return ch.Chat.from_conversation([{"role": "user", "content": message}])

    return self.chat



from fastapi import FastAPI, Request, HTTPException
from fastapi.responses import StreamingResponse, JSONResponse
import httpx
import asyncio
from datetime import datetime

app = FastAPI()

TARGET_URL = "http://nexa:8000"

@app.api_route("/health", methods=["GET"])
def health():
  return JSONResponse(content={"status": "ok"})


@app.api_route("/v1/models", methods=["GET"])
def get_models():
  return JSONResponse(
    content={
      "object": "list",
      "data":
        [{
          "id": "nexa",
          "created": int(datetime.now().timestamp()),
          "object": "model",
          "owned_by": "nexa"
        }]
    }
  )


@app.api_route(
  "/{path:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS"]
)
async def proxy(request: Request, path: str):
  client = httpx.AsyncClient()

  # Construct the target URL
  url = f"{TARGET_URL}/{path}"

  # Forward the request headers
  headers = dict(request.headers)
  headers.pop("host", None)

  try:
    # Stream the request body
    async def request_stream():
      async for chunk in request.stream():
        yield chunk

    # Make the request to the target server
    response = await client.request(
      method=request.method,
      url=url,
      headers=headers,
      params=request.query_params,
      content=request_stream(),
      timeout=httpx.Timeout(300.0)
    )

    # Stream the response back to the client
    async def response_stream():
      try:
          async for chunk in response.aiter_bytes():
              yield chunk
      except (RequestError, ReadTimeout, ConnectTimeout) as e:
          logger.error(f"Error during streaming: {str(e)}")
          # Handle the error appropriately

    return StreamingResponse(
      response_stream(),
      status_code=response.status_code,
      headers=dict(response.headers)
    )

  except httpx.RequestError as exc:
    raise HTTPException(
      status_code=500,
      detail=f"Error communicating with target server: {str(exc)}"
    )

  finally:
    await client.aclose()


if __name__ == "__main__":
  import uvicorn
  uvicorn.run(app, host="0.0.0.0", port=8000)



# klmbr - Kalambur
# https://github.com/av/klmbr

import random

from config import KLMBR_MODS, KLMBR_PERCENTAGE, KLMBR_STRAT, KLMBR_STRAT_PARAMS

import log
import llm
import chat as ch
import selection

logger = log.setup_logger(__name__)

ID_PREFIX = "klmbr"

leetspeak_map = {
  "a": "4",
  "e": "3",
  "i": "1",
  "o": "0",
  "s": "5",
  "t": "7",
  "b": "8",
  "g": "9",
  "l": "1",
}

invert_map = {
  'a': 'ɐ',
  'b': 'q',
  'c': 'ɔ',
  'd': 'p',
  'e': 'ǝ',
  'f': 'ɟ',
  'g': 'ƃ',
  'h': 'ɥ',
  'i': 'ᴉ',
  'j': 'ɾ',
  'k': 'ʞ',
  'l': 'l',
  'n': 'u',
  'o': 'o',
  'p': 'd',
  'q': 'b',
  'r': 'ɹ',
  's': 's',
  't': 'ʇ',
  'u': 'n',
  'v': 'ʌ',
  'w': 'ʍ',
  'x': 'x',
  'y': 'ʎ',
  'z': 'z',
  'A': '∀',
  'B': '𐐒',
  'C': 'Ɔ',
  'D': 'ᗡ',
  'E': 'Ǝ',
  'F': 'Ⅎ',
  'G': '⅁',
  'H': 'H',
  'I': 'I',
  'K': '⋊',
  'L': '˥',
  'M': 'W',
  'N': 'N',
  'O': 'O',
  'P': 'Ԁ',
  'Q': 'Q',
  'R': 'ᴚ',
  'S': 'S',
  'T': '⊥',
  'U': '∩',
  'V': 'Λ',
  'W': 'M',
  'X': 'X',
  'Y': '⅄',
  'Z': 'Z',
}

diacritics = ["̀", "́", "̂", "̃", "̈", "̄", "̆", "̇", "̊", "̋"]

punctuation = ".,!?;:"


def capitalize(chars, idx):
  return chars[idx].swapcase()


def diacritic(chars, idx):
  if chars[idx].isalpha():
    return chars[idx] + random.choice(diacritics)
  return chars[idx]


def leetspeak(chars, idx):
  return leetspeak_map.get(chars[idx].lower(), chars[idx])


def remove_vowel(chars, idx):
  if chars[idx].lower() in "aeiou":
    return ""
  return chars[idx]


def invert_180(chars, idx):
  return invert_map.get(chars[idx], chars[idx])


mods = {
  "capitalize": capitalize,
  "diacritic": diacritic,
  "leetspeak": leetspeak,
  "remove_vowel": remove_vowel,
  "invert_180": invert_180,
}


def modify_text(**kwargs):
  text = kwargs.get("text", "")
  percentage = kwargs.get("percentage", 0)
  target_mods = kwargs.get("mods")

  if target_mods[0] == "all":
    target_mods = list(mods.keys())

  if not text:
    return "", {}

  if not 0 <= percentage <= 100:
    raise ValueError("Percentage must be between 0 and 100")

  words = text.split()
  chars = list(text)
  num_chars_to_modify = max(1, int(len(chars) * (percentage / 100)))
  indices_to_modify = random.sample(range(len(chars)), num_chars_to_modify)
  word_mapping = {}

  for idx in indices_to_modify:
    modification = random.choice(target_mods)

    current_length = 0
    for word_idx, word in enumerate(words):
      if current_length <= idx < current_length + len(word):
        original_word = word
        word_start_idx = current_length
        break
      current_length += len(word) + 1
    else:
      continue

    chars[idx] = mods[modification](chars, idx)
    modified_word = "".join(
      chars[word_start_idx:word_start_idx + len(original_word)]
    )

    if modified_word != original_word:
      cleaned_modified_word = modified_word.rstrip(punctuation)
      cleaned_original_word = original_word.rstrip(punctuation)
      word_mapping[cleaned_modified_word] = cleaned_original_word

  modified_text = "".join(chars)
  return modified_text, word_mapping


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  strat = KLMBR_STRAT.value
  strat_params = KLMBR_STRAT_PARAMS.value
  percentage = KLMBR_PERCENTAGE.value
  mods = KLMBR_MODS.value
  debug_info = {
    "strat": strat,
    "strat_params": strat_params,
    "percentage": percentage,
    "mods": mods,
  }

  logger.debug(f"{ID_PREFIX}: {debug_info}")

  nodes = selection.apply_strategy(chat, strategy=strat, params=strat_params)

  for node in nodes:
    content, mapping = modify_text(
      text=node.content, percentage=percentage, mods=mods
    )
    node.content = content
    node.meta[ID_PREFIX] = mapping

  await llm.emit_status(llm.chat.tail.content)
  await llm.stream_final_completion(chat=chat)



# g1 - the approach from: https://github.com/bklieger-groq/g1
# Harbor also uses same logic for ol1 service

from config import G1_STRAT, G1_STRAT_PARAMS, G1_MAX_STEPS

import llm
import log
import selection
import chat as ch

logger = log.setup_logger(__name__)

ID_PREFIX = "g1"


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  strat = G1_STRAT.value
  strat_params = G1_STRAT_PARAMS.value
  max_steps = G1_MAX_STEPS.value
  debug_info = {
    "strat": strat,
    "strat_params": strat_params,
    "max_steps": max_steps,
  }

  logger.debug(f"{ID_PREFIX}: {debug_info}")

  nodes = selection.apply_strategy(chat, strategy=strat, params=strat_params)

  if (len(nodes) > 1):
    logger.warning(
      f"{ID_PREFIX}: Matched multiple nodes, only the first one will be processed."
    )

  if len(nodes) == 0:
    log.info(f"{ID_PREFIX}: No nodes matched, skipping.")
    return await llm.stream_final_completion()

  node = nodes[0]

  output = ch.Chat(
    llm=llm,
    tail=ch.ChatNode(
      role="system",
      content=
      f"""You are an expert AI assistant that explains your reasoning step by step. For each step, provide a title that describes what you're doing in that step, along with the content. Decide if you need another step or if you're ready to give the final answer. In your response write "ACTION" followed by either 'continue' or 'final_answer'. USE AS MANY REASONING STEPS AS POSSIBLE. AT LEAST 3. BE AWARE OF YOUR LIMITATIONS AS AN LLM AND WHAT YOU CAN AND CANNOT DO. IN YOUR REASONING, INCLUDE EXPLORATION OF ALTERNATIVE ANSWERS. CONSIDER YOU MAY BE WRONG, AND IF YOU ARE WRONG IN YOUR REASONING, WHERE IT WOULD BE. FULLY TEST ALL OTHER POSSIBILITIES. YOU CAN BE WRONG. WHEN YOU SAY YOU ARE RE-EXAMINING, ACTUALLY RE-EXAMINE, AND USE ANOTHER APPROACH TO DO SO. DO NOT JUST SAY YOU ARE RE-EXAMINING. USE AT LEAST 3 METHODS TO DERIVE THE ANSWER. USE BEST PRACTICES."""
      .strip()
    )
  )
  output.user(node.content)
  output.assistant(
    "Thank you! I will now think step by step following my instructions, starting at the beginning after decomposing the problem."
  )

  steps = 0
  while True:
    await llm.emit_status(f'Step: {steps + 1}')
    await output.emit_advance()
    steps += 1

    if output.tail.contains("final_answer") or steps >= max_steps:
      break

  output.user(
    "Please provide the final answer based on your reasoning above. You don't have to mention 'ACTION' in your response."
  )
  await llm.emit_status('Final Answer')
  await llm.stream_final_completion(chat=output)



import logging

def setup_logger(name):
  logger = logging.getLogger(name)
  if not logger.handlers:
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.set_name(name)
    formatter = logging.Formatter(
      "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
  return logger


import re
import requests

url_regex = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
prompt = """
<instruction>
Your task is to fulfill the user's request by discussing provided content.
</instruction>

<content>
{content}
</content>

<request>
{request}
</request>
""".strip()

ID_PREFIX = "discussurl"


async def apply(chat, llm):
  text = chat.text()
  urls = re.findall(url_regex, text)

  # No - URLs - proceed as usual
  if len(urls) == 0:
    return await llm.stream_final_completion()

  # Yes - URLs - read them
  content = ""
  for url in urls:
    await llm.emit_status(f"Reading {url[0]}...")
    content += requests.get(url[0]).text

  await llm.stream_final_completion(
    prompt=prompt,
    content=content,
    request=chat.tail.content,
  )



"""
title: artifact
author: av
author_url: https://github.com/av
description: seeding artifact programmatically
version: 0.0.1
"""

from functools import wraps
import time
import logging
import asyncio
import json
import re

from typing import (
  List,
  Optional,
  AsyncGenerator,
  Callable,
  Awaitable,
  Generator,
  Iterator,
)
from open_webui.constants import TASKS
from open_webui.apps.ollama import main as ollama

# ==============================================================================

name = "artf"
max_steps = 10
final_answer = "<final_answer>"

detect_final = [
  "final answer",
  "final answer.",
  "final answer:",
  "FINAL ANSWER",
]

base_prompt = """
You are an expert AI assistant that explains your reasoning step by step.
For each step, write at most two sentences. BE CONCISE, CLEAR, AND SPECIFIC.
YOU WILL BE PENALIZED FOR WRITING TOO MUCH.
Decide if you need another step or if you're ready to give the final answer.
Avoid markdown in your reply, use HTML tags.
In your response write "ACTION" followed by either 'continue' or '{final_answer}'.
USE AS MANY REASONING STEPS AS POSSIBLE. AT LEAST 3.
BE AWARE OF YOUR LIMITATIONS AS AN LLM AND WHAT YOU CAN AND CANNOT DO.
IN YOUR REASONING, INCLUDE EXPLORATION OF ALTERNATIVE ANSWERS.
CONSIDER YOU MAY BE WRONG, AND IF YOU ARE WRONG IN YOUR REASONING, WHERE IT WOULD BE.
FULLY TEST ALL OTHER POSSIBILITIES. YOU CAN BE WRONG.
WHEN YOU SAY YOU ARE RE-EXAMINING, ACTUALLY RE-EXAMINE, AND USE ANOTHER APPROACH TO DO SO.
DO NOT JUST SAY YOU ARE RE-EXAMINING. USE AT LEAST 3 METHODS TO DERIVE THE ANSWER. USE BEST PRACTICES.
""".strip().format(final_answer=final_answer, max_steps=max_steps)

# ==============================================================================

def throttle(wait_ms):
    def decorator(func):
        last_call = 0
        task = None

        @wraps(func)
        async def wrapper(*args, **kwargs):
            nonlocal last_call, task
            now = time.time()

            if now - last_call >= wait_ms / 1000:
                last_call = now
                if task:
                    task.cancel()
                task = asyncio.create_task(func(*args, **kwargs))
                return await task
            elif task:
                return await task
            else:
                return None

        return wrapper
    return decorator

def is_final_answer(message: str) -> bool:
  return final_answer in message or any([word in message.lower() for word in detect_final])

def setup_logger():
  logger = logging.getLogger(__name__)
  if not logger.handlers:
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.set_name(name)
    formatter = logging.Formatter(
      "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
  return logger


logger = setup_logger()


class Content:
  data: dict

  def __init__(self):
    self.data = {}

  def set(self, key: str, value: str):
    self.data[key] = value
    return self

  def add(self, key: str, value: str):
    if key in self.data:
      self.data[key] += value
    else:
      self.data[key] = value

    return self

  def add_word(self, key: str, value: str):
    self.add(key, f" {value}")
    return self

  def add_tag(self, key: str, tag: str, value: str):
    self.add(key, f"<{tag}>{value}</{tag}>")
    return self

  def add_hr(self, key: str):
    self.add(key, "<hr />")
    return self

  def render(self):
    return f"""
{self.render_artifact()}

---

{self.render_message()}
""".strip()

  def render_message(self):
    return f"""
{self.data.get("message_content", "")}
""".strip()

  # Make a one-liner artifact
  def render_html(self, html: str):
    pattern = r'\s+|(?<=\>)\s+(?=\<)'
    replacement = ' '
    cleaned_html = re.sub(pattern, replacement, html)
    return f"""
```html
{cleaned_html}
```
"""

  def render_artifact(self):
    content = self.data.get("artifact_content", "")
    html = f"""
<html lang="en">
  <head>
    <title>🔮</title>
    <style>
      body {{
        font-family: ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji",Segoe UI Symbol,"Noto Color Emoji";
        background-color: transparent;
        color: #ececec;
        padding: 1rem;
      }}
    </style>
  </head>
  <body>
    <h1>🔮</h1>
    {content}

    <div id="bottom"></div>
    <script>
      const bottom = document.getElementById("bottom");
      bottom.scrollIntoView();
    </script>
  </body>
</html>
    """

    return self.render_html(html)


# ==============================================================================

EventEmitter = Callable[[dict], Awaitable[None]]


class Pipe:
  __current_event_emitter__: EventEmitter
  __question__: str
  __model__: str

  def __init__(self):
    self.type = "manifold"

  def pipes(self) -> list[dict[str, str]]:
    ollama.get_all_models()
    models = ollama.app.state.MODELS

    out = [
      {
        "id": f"{name}-{key}",
        "name": f"{name} {models[key]['name']}"
      } for key in models
    ]
    logger.debug(f"Available models: {out}")

    return out

  def resolve_model(self, body: dict) -> str:
    model_id = body.get("model")
    without_pipe = ".".join(model_id.split(".")[1:])
    return without_pipe.replace(f"{name}-", "")

  def resolve_question(self, body: dict) -> str:
    return body.get("messages")[-1].get("content").strip()

  async def pipe(
    self,
    body: dict,
    __user__: dict,
    __event_emitter__=None,
    __task__=None,
    __model__=None,
  ) -> str | Generator | Iterator:
    model = self.resolve_model(body)
    base_question = self.resolve_question(body)

    if __task__ == TASKS.TITLE_GENERATION:
      content = await self.get_completion(model, body.get("messages"))
      return f"{name}: {content}"

    logger.debug(f"Pipe {name} received: {body}")

    # TODO: concurrency
    self.__model__ = model
    self.__question__ = base_question
    self.__current_event_emitter__ = __event_emitter__

    content = Content()
    content.add_tag("artifact_content", "h3",
                    base_question).add_hr("artifact_content")
    content.set("message_content", '...')

    await self.emit_status("info", "Thinking...", False)

    messages = [
      {
        "role": "system",
        "content": base_prompt
      }, {
        "role": "user",
        "content": base_question
      }, {
        "role":
          "assistant",
        "content":
          "Thank you! I will now think step by step following my instructions, starting at the beginning after decomposing the problem."
      }, {
        "role": "user",
        "content": "Provide the first step."
      }
    ]

    steps = 0
    last_message = ""

    while True:
      if last_message == "":
        content.add_tag("artifact_content", "h3", f"Step: {steps + 1}")

      async for chunk in self.get_streaming_completion(messages):
        last_message += chunk
        content.add("artifact_content", chunk)
        await self.emit_replace(content.render())

      # Smaller models can end generating content early
      # if last_message.strip() != "":
      steps += 1
      messages.append({"role": "assistant", "content": f"\n{last_message}\n"})

      if is_final_answer(last_message) or steps >= max_steps:
        break
      else:
        messages.append({"role": "user", "content": f"Provide the next step. You have {max_steps - steps} remaining."})

      last_message = ""

    content.set("message_content", "\n\n")

    logger.debug(f"Final step: messages={len(messages)}, chat={content.render_message()}")

    messages.append({
      "role": "user",
      "content": f"Now, provide the final answer based on your reasoning above. You don't have to mention {final_answer} in your response anymore."
    })

    await self.emit_replace(content.render())
    await self.emit_status("info", "Final answer.", False)

    async for chunk in self.get_streaming_completion(messages):
      await self.emit_message(chunk)

    await self.done()

  async def progress(
    self,
    message: str,
  ):
    logger.debug(f"Progress: {message}")
    await self.emit_status("info", message, False)

  async def done(self,):
    await self.emit_status("info", "Fin.", True)

  async def emit_message(self, message: str):
    await self.__current_event_emitter__(
      {
        "type": "message",
        "data": {
          "content": message
        }
      }
    )

  @throttle(125)
  async def emit_replace(self, message: str):
    await self.__current_event_emitter__(
      {
        "type": "replace",
        "data": {
          "content": message
        }
      }
    )

  async def emit_status(self, level: str, message: str, done: bool):
    await self.__current_event_emitter__(
      {
        "type": "status",
        "data":
          {
            "status": "complete" if done else "in_progress",
            "level": level,
            "description": message,
            "done": done,
          },
      }
    )

  async def get_streaming_completion(
    self,
    messages,
  ) -> AsyncGenerator[str, None]:
    response = await ollama.generate_openai_chat_completion(
      {
        "model": self.__model__,
        "messages": messages,
        "stream": True
      }
    )

    async for chunk in response.body_iterator:
      for part in self.get_chunk_content(chunk):
        yield part

  async def get_message_completion(self, model: str, content):
    async for chunk in self.get_streaming_completion(
      [{
        "role": "user",
        "content": content
      }]
    ):
      yield chunk

  async def get_completion(self, model: str, messages):
    response = await ollama.generate_openai_chat_completion(
      {
        "model": model,
        "messages": messages,
        "stream": False
      }
    )

    return self.get_response_content(response)

  async def stream_prompt_completion(self, prompt, **format_args):
    complete = ""
    async for chunk in self.get_message_completion(
      self.__model__,
      prompt.format(**format_args),
    ):
      complete += chunk
      await self.emit_message(chunk)
    return complete

  def get_response_content(self, response):
    try:
      return response["choices"][0]["message"]["content"]
    except (KeyError, IndexError):
      logger.error(
        f"ResponseError: unable to extract content from \"{response[:100]}\""
      )
      return ""

  def get_chunk_content(self, chunk):
    chunk_str = chunk.decode("utf-8")
    if chunk_str.startswith("data: "):
      chunk_str = chunk_str[6:]

    chunk_str = chunk_str.strip()

    if chunk_str == "[DONE]" or not chunk_str:
      return

    try:
      chunk_data = json.loads(chunk_str)
      if "choices" in chunk_data and len(chunk_data["choices"]) > 0:
        delta = chunk_data["choices"][0].get("delta", {})
        if "content" in delta:
          yield delta["content"]
    except json.JSONDecodeError:
      logger.error(f"ChunkDecodeError: unable to parse \"{chunk_str[:100]}\"")



import os
import yaml
import argparse
import re

def read_yaml(file_path):
    with open(file_path, 'r') as file:
        return yaml.safe_load(file)

def write_yaml(data, file_path):
    with open(file_path, 'w') as file:
        yaml.dump(data, file, default_flow_style=False)

def render_env_vars(value):
    if isinstance(value, str):
        pattern = r'\$\{([^}]+)\}|\$([a-zA-Z_][a-zA-Z0-9_]*)'

        def replace_env_var(match):
            var_name = match.group(1) or match.group(2)
            return os.environ.get(var_name, match.group(0))

        return re.sub(pattern, replace_env_var, value)
    elif isinstance(value, list):
        return [render_env_vars(item) for item in value]
    elif isinstance(value, dict):
        return {k: render_env_vars(v) for k, v in value.items()}
    else:
        return value

def merge_dicts(dict1, dict2):
    """
    Recursively merge two dictionaries.
    Lists are combined, dictionaries are recursively merged, other values are overwritten.
    """
    result = dict1.copy()
    for key, value in dict2.items():
        if key in result:
            if isinstance(result[key], dict) and isinstance(value, dict):
                result[key] = merge_dicts(result[key], value)
            elif isinstance(result[key], list) and isinstance(value, list):
                result[key].extend(value)
            else:
                result[key] = value
        else:
            result[key] = value
    return result

def merge_yaml_files(directory, pattern, output_file):
    merged_data = {}

    for filename in sorted(os.listdir(directory)):
        if filename.endswith(pattern):
            file_path = os.path.join(directory, filename)
            yaml_data = read_yaml(file_path)

            # Render environment variables
            yaml_data = render_env_vars(yaml_data)

            # Merge the data
            merged_data = merge_dicts(merged_data, yaml_data)

    # Write the merged data to the output file
    write_yaml(merged_data, output_file)

def main():
    parser = argparse.ArgumentParser(description='Merge YAML files in a directory and render environment variables.')
    parser.add_argument('--pattern', default='.yaml', help='File pattern to match (default: .yaml)')
    parser.add_argument('--output', default='merged_output.yaml', help='Output file name (default: merged_output.yaml)')
    parser.add_argument('--directory', default='.', help='Directory to search for YAML files (default: current directory)')

    args = parser.parse_args()

    merge_yaml_files(args.directory, args.pattern, args.output)
    if os.environ.get('HARBOR_LOG_LEVEL', '').upper() == 'DEBUG':
        print(f"Merged YAML files matching '{args.pattern}' into '{args.output}' with environment variables rendered")

if __name__ == '__main__':
    main()


# This is an adjusted version of the:
# https://github.com/tcsenpai/ol1
#
# Which is a fork of the original:
# https://github.com/bklieger-groq/g1

import streamlit as st
import json
import time
import requests  # Add this import for making HTTP requests to Ollama
from dotenv import load_dotenv
import os

load_dotenv()

OLLAMA_URL = os.getenv('OLLAMA_URL', 'http://localhost:11434')
OLLAMA_MODEL = os.getenv('OLLAMA_MODEL', 'llama3.1:8b')
OLLAMA_OPTIONS = os.getenv('OLLAMA_OPTIONS', 'num_predict=300,temperature=0.2')


def parse_options(options):
    parsed_options = {}
    for opt in options.split(','):
        k, v = opt.split('=')
        if k.strip() != "stop":
            parsed_options[k.strip()] = float(v.strip()) if '.' in v else int(v.strip())
        else:
            parsed_options[k.strip()] = v.strip()
    return parsed_options


def make_api_call(messages, max_tokens, is_final_answer=False):
    for attempt in range(3):
        try:
            response = requests.post(f"{OLLAMA_URL}/api/chat",
                                     json={
                                         "model": OLLAMA_MODEL,
                                         "messages": messages,
                                         "stream": False,
                                         "format": "json",
                                         "options": parse_options(OLLAMA_OPTIONS),
                                     })
            response.raise_for_status()
            data = json.loads(response.json()["message"]["content"])
            if "title" not in data or "content" not in data:
                raise ValueError("Response JSON is missing 'title' or 'content' fields.")
            return data

        except Exception as e:
            if attempt == 2:
                if is_final_answer:
                    return {
                        "title":
                        "Error",
                        "content":
                        f"Failed to generate final answer after 3 attempts. Error: {str(e)}"
                    }
                else:
                    return {
                        "title": "Error",
                        "content":
                        f"Failed to generate step after 3 attempts. Error: {str(e)}",
                        "next_action": "final_answer"
                    }
            time.sleep(1)  # Wait for 1 second before retrying


def generate_response(prompt):
    messages = [{
        "role":
        "system",
        "content":
        """You are an expert AI assistant that explains your reasoning step by step. For each step, provide a title that describes what you're doing in that step, along with the content. Decide if you need another step or if you're ready to give the final answer. Respond in JSON format with 'title', 'content', and 'next_action' (either 'continue' or 'final_answer') keys. USE AS MANY REASONING STEPS AS POSSIBLE. AT LEAST 3. BE AWARE OF YOUR LIMITATIONS AS AN LLM AND WHAT YOU CAN AND CANNOT DO. IN YOUR REASONING, INCLUDE EXPLORATION OF ALTERNATIVE ANSWERS. CONSIDER YOU MAY BE WRONG, AND IF YOU ARE WRONG IN YOUR REASONING, WHERE IT WOULD BE. FULLY TEST ALL OTHER POSSIBILITIES. YOU CAN BE WRONG. WHEN YOU SAY YOU ARE RE-EXAMINING, ACTUALLY RE-EXAMINE, AND USE ANOTHER APPROACH TO DO SO. DO NOT JUST SAY YOU ARE RE-EXAMINING. USE AT LEAST 3 METHODS TO DERIVE THE ANSWER. USE BEST PRACTICES.

Example of a valid JSON response:
```json
{
    "title": "Identifying Key Information",
    "content": "To begin solving this problem, we need to carefully examine the given information and identify the crucial elements that will guide our solution process. This involves...",
    "next_action": "continue"
}```
"""
    }, {
        "role": "user",
        "content": prompt
    }, {
        "role":
        "assistant",
        "content":
        "Thank you! I will now think step by step following my instructions, starting at the beginning after decomposing the problem."
    }]

    steps = []
    step_count = 1
    total_thinking_time = 0

    while True:
        start_time = time.time()
        step_data = make_api_call(messages, 300)
        end_time = time.time()
        thinking_time = end_time - start_time
        total_thinking_time += thinking_time

        steps.append(
            (f"Step {step_count}: {step_data.get('title', 'Untitled Step')}",
             step_data.get('content', json.dumps(step_data,
                                                 indent=2)), thinking_time))

        messages.append({
            "role": "assistant",
            "content": json.dumps(step_data)
        })

        if step_data.get('next_action', 'continue') == 'final_answer':
            break

        step_count += 1

        # Yield after each step for Streamlit to update
        yield steps, None  # We're not yielding the total time until the end

    # Generate final answer
    messages.append({
        "role":
        "user",
        "content":
        "Please provide the final answer based on your reasoning above."
    })

    start_time = time.time()
    final_data = make_api_call(messages, 200, is_final_answer=True)
    end_time = time.time()
    thinking_time = end_time - start_time
    total_thinking_time += thinking_time

    steps.append(("Final Answer", final_data['content'], thinking_time))

    yield steps, total_thinking_time


def main():
    st.set_page_config(page_title="ol1", page_icon="🧠", layout="wide")

    st.markdown(f"**Current Configuration:**")
    st.markdown(f"- Ollama URL: `{OLLAMA_URL}`")
    st.markdown(f"- Ollama Model: `{OLLAMA_MODEL}`")

    # Initialize session state
    if 'has_generated_response' not in st.session_state:
        st.session_state.has_generated_response = False

    # Text input for user query
    user_query = st.text_input(
        "Enter your query:",
        placeholder="e.g., How many 'R's are in the word strawberry?")

    if user_query and not st.session_state.has_generated_response:
        st.write("Generating response...")
        # Create empty elements to hold the generated text and total time
        response_container = st.empty()
        time_container = st.empty()

        # Generate and display the response
        for steps, total_thinking_time in generate_response(user_query):
            with response_container.container():
                for i, (title, content, thinking_time) in enumerate(steps):
                    if title.startswith("Final Answer"):
                        st.markdown(f"### {title}")
                        st.markdown(content.replace('\n', '<br>'),
                                    unsafe_allow_html=True)
                    else:
                        with st.expander(title, expanded=True):
                            st.markdown(content.replace('\n', '<br>'),
                                        unsafe_allow_html=True)
            # Only show total time when it's available at the end
            if total_thinking_time is not None:
                time_container.markdown(
                    f"**Total thinking time: {total_thinking_time:.2f} seconds**"
                )

        st.session_state.has_generated_response = True

    # Retry button
    if st.session_state.has_generated_response:
        if st.button("Retry"):
            st.session_state.has_generated_response = False
            st.session_state.user_query = ""
            st.rerun()


if __name__ == "__main__":
    main()



# Adjusted from
# https://github.com/fedirz/parler-tts-server/blob/master/parler_tts_server/main.py
# With:

import time
from contextlib import asynccontextmanager
from typing import Annotated, Any, OrderedDict

import huggingface_hub
import soundfile as sf
import torch
from fastapi import Body, FastAPI, HTTPException, Response
from fastapi.responses import FileResponse
from huggingface_hub.hf_api import ModelInfo
from openai.types import Model
from parler_tts import ParlerTTSForConditionalGeneration
from transformers import AutoTokenizer

from parler_tts_server.config import SPEED, ResponseFormat, config
from parler_tts_server.logger import logger

# https://github.com/huggingface/parler-tts?tab=readme-ov-file#usage
if torch.cuda.is_available():
    device = "cuda:0"
    logger.info("GPU will be used for inference")
else:
    device = "cpu"
    logger.info("CPU will be used for inference")
torch_dtype = torch.float16 if device != "cpu" else torch.float32

class ModelManager:
    def __init__(self):
        self.model_tokenizer: OrderedDict[
            str, tuple[ParlerTTSForConditionalGeneration, AutoTokenizer]
        ] = OrderedDict()

    def load_model(
        self, model_name: str
    ) -> tuple[ParlerTTSForConditionalGeneration, AutoTokenizer]:
        logger.debug(f"Loading {model_name}...")
        start = time.perf_counter()

        tokenizer = AutoTokenizer.from_pretrained(model_name)
        model = ParlerTTSForConditionalGeneration.from_pretrained(
            model_name
          ).to(  # type: ignore
            device,  # type: ignore
            dtype=torch_dtype,
        )

        # Harbor Patch:
        # https://github.com/huggingface/parler-tts/blob/main/INFERENCE.md
        # TODO: PyTorch 2.4 required to compile with Python 3.12
        # compile_mode = "default" # chose "reduce-overhead" for 3 to 4x speed-up
        # logger.info(
        #     f"Compiling with the mode {compile_mode}..."
        # )
        # model.generation_config.cache_implementation = "static"
        # model.forward = torch.compile(model.forward, mode=compile_mode)
        # warmup
        # inputs = tokenizer("This is for compilation", return_tensors="pt", padding="max_length", max_length=max_length).to(device)
        # model_kwargs = {**inputs, "prompt_input_ids": inputs.input_ids, "prompt_attention_mask": inputs.attention_mask, }
        # n_steps = 1 if compile_mode == "default" else 2
        # for _ in range(n_steps):
        #     _ = model.generate(**model_kwargs)

        logger.info(
            f"Loaded {model_name} and tokenizer in {time.perf_counter() - start:.2f} seconds"
        )
        return model, tokenizer

    def get_or_load_model(
        self, model_name: str
    ) -> tuple[ParlerTTSForConditionalGeneration, Any]:
        if model_name not in self.model_tokenizer:
            logger.info(f"Model {model_name} isn't already loaded")
            if len(self.model_tokenizer) == config.max_models:
                logger.info("Unloading the oldest loaded model")
                del self.model_tokenizer[next(iter(self.model_tokenizer))]
            self.model_tokenizer[model_name] = self.load_model(model_name)
        return self.model_tokenizer[model_name]


model_manager = ModelManager()


@asynccontextmanager
async def lifespan(_: FastAPI):
    if not config.lazy_load_model:
        model_manager.get_or_load_model(config.model)
    yield


app = FastAPI(lifespan=lifespan)


@app.get("/health")
def health() -> Response:
    return Response(status_code=200, content="OK")


@app.get("/v1/models", response_model=list[Model])
def get_models() -> list[Model]:
    models = list(huggingface_hub.list_models(model_name="parler-tts"))
    models = [
        Model(
            id=model.id,
            created=int(model.created_at.timestamp()),
            object="model",
            owned_by=model.id.split("/")[0],
        )
        for model in models
        if model.created_at is not None
    ]
    return models


@app.get("/v1/models/{model_name:path}", response_model=Model)
def get_model(model_name: str) -> Model:
    models = list(huggingface_hub.list_models(model_name=model_name))
    if len(models) == 0:
        raise HTTPException(status_code=404, detail="Model doesn't exists")
    exact_match: ModelInfo | None = None
    for model in models:
        if model.id == model_name:
            exact_match = model
            break
    if exact_match is None:
        raise HTTPException(
            status_code=404,
            detail=f"Model doesn't exists. Possible matches: {", ".join([model.id for model in models])}",
        )
    assert exact_match.created_at is not None
    return Model(
        id=exact_match.id,
        created=int(exact_match.created_at.timestamp()),
        object="model",
        owned_by=exact_match.id.split("/")[0],
    )


# https://platform.openai.com/docs/api-reference/audio/createSpeech
@app.post("/v1/audio/speech")
async def generate_audio(
    input: Annotated[str, Body()],
    voice: Annotated[str, Body()] = config.voice,
    model: Annotated[str, Body()] = config.model,
    response_format: Annotated[ResponseFormat, Body()] = config.response_format,
    speed: Annotated[float, Body()] = SPEED,
) -> FileResponse:
    tts, tokenizer = model_manager.get_or_load_model(model)
    if speed != SPEED:
        logger.warning(
            "Specifying speed isn't supported by this model. Audio will be generated with the default speed"
        )
    start = time.perf_counter()
    input_ids = tokenizer(voice, return_tensors="pt").input_ids.to(device)
    prompt_input_ids = tokenizer(input, return_tensors="pt").input_ids.to(device)
    generation = tts.generate(
        input_ids=input_ids, prompt_input_ids=prompt_input_ids
    ).to(  # type: ignore
        torch.float32
    )
    audio_arr = generation.cpu().numpy().squeeze()
    logger.info(
        f"Took {time.perf_counter() - start:.2f} seconds to generate audio for {len(input.split())} words using {device.upper()}"
    )
    # TODO: use an in-memory file instead of writing to disk
    sf.write(f"out.{response_format}", audio_arr, tts.config.sampling_rate)
    return FileResponse(f"out.{response_format}", media_type=f"audio/{response_format}")


from config import STATUS_STYLE

status_formatters = {
  "md:codeblock": "\n```boost\n{status}\n```\n",
  "md:h1": "\n\n# {status}\n\n",
  "md:h2": "\n\n## {status}\n\n",
  "md:h3": "\n\n### {status}\n\n",
  "plain": "\n\n{status}\n\n",
  "none": ""
}

def format_status(status: str):
  desired_format = STATUS_STYLE.value

  if desired_format not in status_formatters:
    desired_format = "md:codeblock"

  return status_formatters[desired_format].format(status=status)
